
library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all;
  use ieee.std_logic_arith.all;

entity RB65_tb is

end entity;

architecture tb of RB65_tb is
  signal A	: std_logic_vector(15 downto 0);
  signal A_i	: unsigned(15 downto 0);
  signal A_r	: unsigned(12 downto 0);
  signal Din,
         Dout,
         D,
	 Drom   : std_logic_vector(7 downto 0);
  signal IRQ    : std_logic;
  signal m49	: std_logic := '0';
  signal NMI    : std_logic;
  signal PHI0	: std_logic := '0';
  signal PHI1   : std_logic;
  signal PHI2   : std_logic;
  signal RDY    : std_logic;
  signal reset	: std_logic := '0';
  signal RW     : std_logic;			-- Read/Write line
  signal SO_n   : std_logic;
  signal SSB    : std_logic;			-- Single Step Busy
  signal SSC    : std_logic;			-- Single Step Clock
  signal SYNC   : std_logic;
  


 type RAM_type is array(0 to 65535) of std_logic_vector(7 downto 0);
 signal RAM_table : RAM_type;


  
begin
  i_ROM: entity work.ROM_Table
  port map (
    Ain => A_r,
    Bout => Drom
  );
  
  
  i_RB65: entity work.RB65g
  port map (
    A => A,
    D => D,
    Dout => Dout,
    IRQ => IRQ,
    m49 => m49,
    NMI => NMI,
    PHI0 => PHI0,
    PHI1 => PHI1,
    PHI2 => PHI2,
    RDY => RDY,
    RESET  => reset,
    RW => RW,
    SO_n => SO_n,
    SSB => SSB,
    SSC => SSC,
    SYNC => SYNC
  );
  
  
  
  m49 <= not m49 after 11 ns;
  PHI0 <= not PHI0 after 500 ns;
  reset <= '0', '1' after 800 ns;
  D <= Din when (RW = '1') else Dout;
  A_i <= unsigned(A);
  A_r <= unsigned(A(12 downto 0));
  

  process(PHI0)
  begin
    if rising_edge(PHI0) then
      if (A(15 downto 13) = "111") then
        Din <= Drom;
      else
        Din <= RAM_table(conv_integer(A_i));
     end if;
    end if;
  end process;


  process(PHI0)
  begin
    if (falling_edge(PHI0) and (RW = '0')) then
      RAM_table(conv_integer(A_i)) <= Dout;
    end if;
  end process;

end tb;
















